import asyncio
import json
import aiohttp
from .backend import Vkontakte
from ...logger import logger


class VkontakteLongpoll(Vkontakte):
    def __init__(self, *args, longpoll_settings=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.longpoll_settings = longpoll_settings or {}
        self.longpoll_data = {}
        self.longpoll_url_template = "{}?act=a_check&key={}&wait=25&ts={}"

    async def update_longpoll_data(self):
        self.longpoll_data = await self.raw_request(
            "groups.getLongPollServer", {"group_id": self.group_id}
        )

    async def _get_updates(self):
        longpoll_url = self.longpoll_url_template.format(
            self.longpoll_data["server"],
            self.longpoll_data["key"],
            self.longpoll_data["ts"],
        )

        try:
            async with self.session.post(longpoll_url) as resp:
                return await resp.json(content_type=None)
        except (json.JSONDecodeError, aiohttp.ClientError, asyncio.TimeoutError):
            return None

        except asyncio.CancelledError:
            raise

        except Exception:
            logger.exception("Exceptions while gettings updates (Vkontakte)")
            await asyncio.sleep(1)
            return None

    async def acquire_updates(self, submit_update):
        response = await self._get_updates()

        if response is None:
            return

        if "ts" in response:
            self.longpoll_data["ts"] = response["ts"]

        if "failed" in response:
            if response["failed"] in (2, 3):
                await self.update_longpoll_data()
            return

        for update_data in response["updates"]:
            if "type" not in update_data or "object" not in update_data:
                continue

            if update_data["type"] == "group_change_settings":
                changes = update_data["object"]["changes"]

                screen_name = changes.get("screen_name")
                if screen_name:
                    self.group_screen_name = screen_name["new_value"]

                title = changes.get("title")
                if title:
                    self.group_name = title["new_value"]

            await submit_update(self._make_update(update_data))

    async def on_start(self, app):
        await super().on_start(app)

        longpoll_settings = {
            **self.default_updates_settings,
            **self.longpoll_settings,
        }

        await self.raw_request(
            "groups.setLongPollSettings",
            {
                "group_id": self.group_id,
                "api_version": self.api_version,
                "enabled": 1,
                **longpoll_settings,
            }
        )

        await self.update_longpoll_data()



from urllib.parse import urlparse
import asyncio
import re
from aiohttp import web
from .backend import Vkontakte
from ...helpers import get_random_string
from ...logger import logger


DEFAULT_ADDRESS = "0.0.0.0:8080/callback/1"


class VkontakteCallback(Vkontakte):
    def __init__(
        self,
        *args,
        address=None,
        address_path=None,
        host="0.0.0.0",
        port=10888,
        queue_limit=0,
        callback_settings=None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.callback_settings = callback_settings or {}

        if address:
            if not re.match(r"^https://", address):
                address = f"https://{address}"

            self._address = address
            self._address_path = address_path or urlparse(address).path or "/"

        else:
            self._address = None
            self._address_path = address_path or "/"

        self._server_path = self._address_path
        self._server_host = host
        self._server_port = port
        self._server_app_runner = None

        self._queue_limit = queue_limit
        self.updates_queue = None

    async def handle_request(self, request):
        data = await request.json()

        if data["type"] == "confirmation":
            resp = await self.request(
                "groups.getCallbackConfirmationCode",
                group_id=data["group_id"]
            )

            return web.Response(body=f"{resp['code']}")

        await self.updates_queue.put(self._make_update(data))

        return web.Response(body="ok")

    def make_server_app(self):
        app = web.Application()
        app.add_routes([web.post(self._server_path, self.handle_request)])
        return app

    async def start_server(self):
        app = self.make_server_app()

        self._server_app_runner = web.AppRunner(app)

        await self._server_app_runner.setup()

        site = web.TCPSite(
            self._server_app_runner,
            self._server_host,
            self._server_port
        )

        await site.start()

    async def stop_server(self):
        if self._server_app_runner:
            await self._server_app_runner.cleanup()

    async def on_start(self, app):
        await super().on_start(app)

        # Setup queue
        self.updates_queue = asyncio.Queue(self._queue_limit, loop=app.get_loop())

        # Start web server
        await self.start_server()

        if not self._address:
            logger.warning(
                "No address provided for VkontakteCallback! You "
                "will have to setup your group's settings manually."
            )
            return

        # Delete existing servers in group's callback servers
        servers = await self.request(
            "groups.getCallbackServers",
            group_id=self.group_id
        )

        title_num = 1

        for server in servers["items"]:
            if server["url"] == self._address:
                await self.request(
                    "groups.deleteCallbackServer",
                    group_id=self.group_id,
                    server_id=server["id"],
                )

            elif server["title"].startswith("kutana@") and server["title"][7:].isdigit():
                title_num = max(title_num, int(server["title"][7:]) + 1)

        # Add server to group's callback servers
        response = await self.request(
            "groups.addCallbackServer",
            group_id=self.group_id,
            url=self._address,
            title=f"kutana@{title_num}",
            secret_key=get_random_string(),
        )

        server_id = response["server_id"]

        # Setup callback server
        callback_settings = {
            **self.default_updates_settings,
            **self.callback_settings,
        }

        await self.request(
            "groups.setCallbackSettings",
            group_id=self.group_id,
            api_version=self.api_version,
            server_id=server_id,
            **callback_settings,
        )

    async def on_shutdown(self, app):
        await super().on_shutdown(app)
        await self.stop_server()

    async def acquire_updates(self, submit_update):
        await submit_update(await self.updates_queue.get())



import random
from ..storage import Storage, OptimisticLockException


class MemoryStorage(Storage):
    """
    Naive implementation of the storage that uses in-memory dict for
    storing data.

    WARNING:
        When keys count is equal or greater than 'keys_limit', 30% of the
        stored keys will be cleared.
    """

    def __init__(self, keys_limit=1_000_000):
        self.keys_limit = keys_limit
        self._storage = {}

    async def _put(self, key, values, version=None):
        # handle keys overflow
        if len(self._storage) >= self.keys_limit:
            for key in random.sample(list(self._storage.keys()), k=int(self.keys_limit * 0.3)):
                self._storage.pop(key)

        # use optimistic locking
        if version is None and key in self._storage:
            raise OptimisticLockException("Values for this key already exists")

        if version is not None and key not in self._storage:
            raise OptimisticLockException("Values for this key was deleted")

        if self._storage.get(key, {}).get("_version") != version:
            raise OptimisticLockException("Versions differ from expected value")

        # update value
        new_version = (version or 0) + 1
        self._storage[key] = {**values, "_version": new_version}
        return new_version

    async def _get(self, key):
        return self._storage.get(key, None)

    async def _delete(self, key):
        self._storage.pop(key, None)



import string
import random
import os.path


def uniq_by(arr, key=None):
    if not key:
        def key(element):
            return element
    return list({key(element): element for element in arr}.values())


def pick(obj, keys):
    new_obj = {}
    for key in keys:
        if key in obj:
            new_obj[key] = obj[key]
    return new_obj


def get_random_string(length=16):
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def get_path(root, path):
    """
    Shortcut for ``os.path.join(os.path.dirname(root), path)``.

    :param root: root path
    :param path: path to file or folder
    :returns: path to file or folder relative to root
    """

    return os.path.join(os.path.dirname(root), path)


def ensure_list(value):
    if isinstance(value, (list, tuple)):
        return value
    return [value]



from .helpers import ensure_list


class OptimisticLockException(Exception):
    pass


class Document:
    def __init__(self, values, key=None, storage=None):
        self.key = key
        self.storage = storage
        self.values = values

    @property
    def version(self):
        return self.values.get("_version", None)

    def __setattr__(self, key, value):
        if key in ("values", "key", "storage"):
            super().__setattr__(key, value)
        else:
            self.__dict__["values"][key] = value

    def __getattr__(self, key):
        return self.values[key]

    def __setitem__(self, key, value):
        self.values[key] = value

    def __getitem__(self, key):
        return self.values[key]

    def __len__(self):
        return len(self.values)

    def __delitem__(self, key):
        del self.values[key]

    def __contains__(self, item):
        return item in self.values

    def __iter__(self):
        return iter(self.values)

    def __str__(self):
        return f"Document('{self.key}', {self.values})"

    async def update(self, values, remove=()):
        self.values.update(values)
        for field in ensure_list(remove):
            self.values.pop(field)
        return await self.save()

    async def save(self):
        """
        Save this document. Raises :class:`kutana.storage.OptimisticLockException` on
        issues with optimistic locking.
        """
        if not self.key:
            raise ValueError("This document missing a key")
        if not self.storage:
            raise ValueError("This document missing a storage")

        self.values["_version"] = await self.storage.put(self.key, self.values)

        return self

    async def refresh(self):
        """
        Load current data from the storage and replace current one with it.
        """
        current = await self.storage.get(self.key)
        if not current:
            raise ValueError("This document is deleted now")

        self.values = current.values

        return self


class Storage:
    async def init(self):
        pass

    def make_document(self, values, key):
        return Document({**values, "_version": None}, key=key, storage=self)

    async def put(self, key, values):
        if "_version" not in values:
            raise ValueError("Values missing '_version'")
        return await self._put(key, values, version=values.get("_version"))

    async def delete(self, key):
        return await self._delete(key)

    async def get(self, key):
        values = await self._get(key)
        if not values:
            return None
        return Document(values, key=key, storage=self)

    async def _put(self, key, values, version=None):
        raise NotImplementedError

    async def _get(self, key):
        raise NotImplementedError

    async def _delete(self, key):
        raise NotImplementedError



from .debug import Debug
from .vkontakte import Vkontakte, VkontakteLongpoll, VkontakteCallback
from .telegram import Telegram
from .terminal import Terminal

__all__ = [
    "Debug",
    "Vkontakte", "VkontakteLongpoll", "VkontakteCallback",
    "Telegram",
    "Terminal"
]



import os
import ast
import argparse
import os.path
import yaml


parser = argparse.ArgumentParser("kutana-i18n", description="Run kutana i18n related tasks.")
subparsers = parser.add_subparsers(help="action to perform", dest="action", required=True)

parser_collect = subparsers.add_parser("collect", help="collect translation string from sources")
parser_collect.add_argument(
    "--collect_from", "-c", dest="collect_from",
    default=os.getcwd(), help="path to search for sources"
)
parser_collect.add_argument(
    "--save_to", "-s", dest="save_to",
    default="i18n.yml", help="file to save results to"
)


def get_static_value(node):
    if node and node.__class__.__name__ == "Constant":
        return node.value
    return None


def collect_from_file(path, translation):
    with open(path, "r") as fh:
        content = fh.read()

    alias = ""

    root = ast.parse(content)

    for node in ast.walk(root):
        if node.__class__.__name__ == "ImportFrom":
            if node.module not in ["kutana", "kutana.i18n"]:
                continue

            for name in node.names:
                if name.name == "t":
                    alias = name.asname or "t"

        if (alias and node.__class__.__name__ == "Call"
                and node.func.__class__.__name__ == "Name"
                and node.func.id == alias):
            msgid = get_static_value(node.args[0] if node.args else None)
            known_kwargs = {kw.arg: get_static_value(kw.value) for kw in node.keywords}
            ctx = known_kwargs.get("ctx")
            num = "num" in known_kwargs

            key = (msgid, ctx or "")
            pointer = f"{path}:{node.lineno}:{node.col_offset + 1}"

            if key in translation:
                translation[key]["$where"].append(pointer)
            else:
                translation[key] = {
                    "msgid": msgid,
                    "msgstr": [] if num else "",
                    **({"msgctx": ctx} if ctx else {}),
                    "$where": [pointer],
                }


def collect(path, target):
    translation = {}

    if os.path.isfile(target):
        with open(target, "r") as fh:
            for string in yaml.safe_load(fh.read()):
                key = (string["msgid"], string.get("msgctx") or "")

                translation[key] = {
                    "msgid": string["msgid"],
                    "msgstr": string["msgstr"],
                    **({"msgctx": string["msgctx"]} if string.get("msgctx") else {}),
                    "$where": [],
                }

    for dirpath, __, filenames in os.walk(os.getcwd()):
        for filename in filenames:
            if not filename.endswith(".py"):
                continue

            collect_from_file(
                os.path.relpath(os.path.join(dirpath, filename), os.getcwd()),
                translation
            )

    if os.path.dirname(target):
        os.makedirs(os.path.dirname(target), exist_ok=True)

    with open(target, "w") as fh:
        for translation in sorted([t for t in translation.values() if t["$where"]], key=lambda t: t["$where"]):
            where = translation.pop("$where", [])
            print("# " + "\n# ".join(where), file=fh)
            print(yaml.safe_dump([translation], allow_unicode=True), file=fh)

        if not fh.tell():
            fh.write("[]")


def run():
    args = parser.parse_args()

    if args.action == "collect":
        collect(args.collect_from, args.save_to)
        exit(0)


if __name__ == "__main__":
    run()



import re
import inspect
import warnings
import functools
from .handler import Handler, HandlerResponse
from .update import UpdateType
from .helpers import ensure_list
from .backends.vkontakte import VkontaktePluginExtension
from .routers import (
    CommandsRouter, AttachmentsRouter, ListRouter, AnyMessageRouter,
    AnyUpdateRouter
)


class Plugin:
    """
    Plugin provides functionality to the application.

    :param name: Name of the plugin
    :param description: Description of the plugin
    """

    def __init__(self, name, description="", storage="default", **kwargs):
        self.app = None
        self.name = name
        self.description = description

        self._routers = []
        self._storage = storage
        self._handlers = {
            "start": [],
            "before": [],
            "after": [],
            "exception": [],
            "shutdown": [],
        }

        # Set provided attributes to this instance
        for k, v in kwargs.items():
            setattr(self, k, v)

    def _get_or_add_router(self, router_cls, priority=None):
        for router in self._routers:
            if type(router) == router_cls and (priority is None or priority == router.priority):
                return router

        if priority is None:
            router = router_cls()
        else:
            router = router_cls(priority=priority)

        self._routers.append(router)
        return router

    def _add_handler_for_router(self, router, handler, handler_key=None, router_priority=None):
        router = self._get_or_add_router(router, priority=router_priority)
        if handler_key:
            router.add_handler(handler, handler_key)
        else:
            router.add_handler(handler)

    @property
    def storage(self):
        if isinstance(self._storage, str):
            return self.app.get_storage(self._storage)
        return self._storage

    def on_start(self, priority=0):
        """
        Decorator for registering coroutine to be called when application
        starts.

        Priority specifies the order in which handlers are  executed. Handlers
        with higher priority will be executed first.
        """
        def decorator(func):
            async def wrapper():
                args, *__ = inspect.getfullargspec(func)

                if args:
                    warnings.warn(
                        '"on_start" with arguments is deprecated, you can '
                        'access "app" with "plugin.app"',
                        DeprecationWarning
                    )
                    return await func(self.app)
                else:
                    return await func()

            self._handlers["start"].append(Handler(wrapper, priority))

            return func
        return decorator

    def on_before(self, priority=0):
        """
        Decorator for registering coroutine to be called before an update
        will be processed. It will be passed the update and it's context as
        arguments.

        Priority specifies the order in which handlers are executed. Handlers
        with higher priority will be executed first.
        """
        def decorator(func):
            self._handlers["before"].append(Handler(func, priority))
            return func
        return decorator

    def on_after(self, priority=0):
        """
        Decorator for registering coroutine to be called after an update
        was processed. It will be passed the update, it's context and
        result of processing (:class:`kutana.handler.HandlerResponse`) as
        arguments.

        Priority specifies the order in which handlers are executed. Handlers
        with higher priority will be executed first.
        """
        def decorator(func):
            self._handlers["after"].append(Handler(func, priority))
            return func
        return decorator

    def on_exception(self, priority=0):
        """
        Decorator for registering coroutine to be called when exception was
        raised while processing an update. It will be passed the update,
        it's context and raised exception as arguments.

        Priority specifies the order in which handlers are executed. Handlers
        with higher priority will be executed first.
        """
        def decorator(func):
            self._handlers["exception"].append(Handler(func, priority))
            return func
        return decorator

    def on_shutdown(self, priority=0):
        """
        Decorator for registering coroutine to be called when application
        will be stopped. It will be passed application as an argument.

        Priority specifies the order in which handlers are executed. Handlers
        with higher priority will be executed first.
        """
        def decorator(func):
            self._handlers["shutdown"].append(Handler(func, priority))
            return func
        return decorator

    # Registrators for updates with specific conditions
    @property
    def vk(self) -> VkontaktePluginExtension:
        return VkontaktePluginExtension(self)

    def on_commands(
        self,
        commands,
        priority=0,
        router_priority=None
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message, starts with prefix and one of
        provided commands.

        If case_insensitive is True, commands will be processed with ignored
        case.

        Context is automatically populated with following values:

        - prefix
        - command
        - body
        - match

        Priority is a value that used to determine order in which
        coroutines are executed. Router priority is a value that
        used to determine the order in which routers is checked
        for appropriate handler. By default, most of the times,
        priority is 0, and router_priority is None (default
        value will be used).

        If coroutine returns anythign but
        :class:`kutana.handler.HandlerResponse` SKIPPED, other
        coroutines are not executed further.
        """

        commands = ensure_list(commands)

        def decorator(func):
            for command in commands:
                self._add_handler_for_router(
                    CommandsRouter,
                    handler=Handler(func, priority),
                    handler_key=command,
                    router_priority=router_priority,
                )
            return func

        return decorator

    def on_attachments(
        self,
        types,
        priority=0,
        router_priority=None
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message and have at least one
        attachment with one of specified types.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        types = ensure_list(types)

        def decorator(func):
            for atype in types:
                self._add_handler_for_router(
                    AttachmentsRouter,
                    handler=Handler(func, priority),
                    handler_key=atype,
                    router_priority=router_priority,
                )
            return func

        return decorator

    def on_any_message(self, *args, **kwargs):
        warnings.warn(
            '"on_any_message" is deprecated, use "on_messages" instead',
            DeprecationWarning
        )
        return self.on_messages(*args, **kwargs)

    def _make_decorator(self, router, priority, router_priority):
        def decorator(func):
            self._add_handler_for_router(
                router,
                handler=Handler(func, priority),
                router_priority=router_priority,
            )
            return func

        return decorator

    def on_messages(self, priority=0, router_priority=None):
        """
        Decorator for registering coroutine to be called when
        incoming update is message. This will always be called.
        If you want to catch all unprocessed message, you should
        use :meth:`kutana.plugin.Plugin.on_unprocessed_messages`.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        return self._make_decorator(AnyMessageRouter, priority, router_priority)

    def on_any_unprocessed_message(self, *args, **kwargs):
        warnings.warn(
            '"on_any_unprocessed_message" is deprecated, use "on_unprocessed_messages" instead',
            DeprecationWarning
        )
        return self.on_unprocessed_messages(*args, **kwargs)

    def on_unprocessed_messages(
        self,
        priority=0,
        router_priority=-3
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message. This will be called if no other
        coroutine processed update.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        return self._make_decorator(AnyMessageRouter, priority, router_priority)

    def on_match(self, pattern, priority=0, router_priority=-3):
        """
        Decorator for registering coroutine to be called when
        incoming update is message and it's text is matched by
        provided pattern.

        Context is automatically populated with following values:

        - match

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        def decorator(func):
            async def wrapper(update, ctx):
                if update.type != UpdateType.MSG:
                    return HandlerResponse.SKIPPED

                match = re.match(pattern, update.text)
                if not match:
                    return HandlerResponse.SKIPPED

                ctx.match = match

                return await func(update, ctx)

            self._add_handler_for_router(
                ListRouter,
                handler=Handler(wrapper, priority),
                router_priority=router_priority,
            )

            return func

        return decorator

    def on_any_update(self, *args, **kwargs):
        warnings.warn(
            '"on_any_update" is deprecated, use "on_updates" instead',
            DeprecationWarning
        )
        return self.on_updates(self, *args, **kwargs)

    def on_updates(self, priority=0, router_priority=None):
        """
        Decorator for registering coroutine to be called when
        incoming update is not message. This will always be called.
        If you want to catch all unprocessed message, you should
        use :meth:`kutana.plugin.Plugin.on_unprocessed_updates`.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        return self._make_decorator(AnyUpdateRouter, priority, router_priority)

    def on_any_unprocessed_update(self, *args, **kwargs):
        warnings.warn(
            '"on_any_unprocessed_update" is deprecated, use "on_unprocessed_updates" instead',
            DeprecationWarning
        )
        return self.on_unprocessed_updates(self, *args, **kwargs)

    def on_unprocessed_updates(self, priority=0, router_priority=-3):
        """
        Decorator for registering coroutine to be called when
        incoming update is update. This will be called if no other
        coroutine processed update.

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        return self._make_decorator(AnyUpdateRouter, priority, router_priority)

    def expect_sender(self, state=None, localized=False):
        """
        This decorator does following things:

        - It adds 'sender' to the context. Object that contains data you
            previously saved to it. You can treat it as object or dict
            with data. In order to save your changes, you should await
            'save' method (or use 'update').

        - If you specified 'state', this will skip all users that don't have
            specified value in 'state' field.

        - If you 'localized' is true, state will be limited only to current
            receiver.

        NOTE:
            If you will attempt to update state, but between reading and
            writing (updating) someone already updated it 'OptimisticLockException'
            will be raised. By default this exceptions are ignored!
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(upd, ctx, *args, **kwargs):
                sender = await self.storage.get(ctx.sender_key)

                if not sender or "state" not in sender:
                    sender = self.storage.make_document(
                        {"state": "", "_version": None}, ctx.sender_key
                    )

                if state is not None and sender["state"] != state:
                    return HandlerResponse.SKIPPED

                ctx.sender = sender

                return await func(upd, ctx, *args, **kwargs)
            return wrapper
        return decorator

    def expect_receiver(self, state=None):
        """
        This decorator does following things:

        - It adds 'receiver' to the context. Object that contains data you
            previously saved to it. You can treat it as object or dict
            with data. In order to save your changes, you should await
            'save' method (or use 'update').

        NOTE:
            If you will attempt to update state, but between reading and
            writing (updating) someone already updated it - 'OptimisticLockException'
            will be raised. By default this exceptions are ignored!
        """

        def decorator(func):
            @functools.wraps(func)
            async def wrapper(upd, ctx, *args, **kwargs):
                receiver = await self.storage.get(ctx.receiver_key)

                if not receiver or "state" not in receiver:
                    receiver = self.storage.make_document(
                        {"state": "", "_version": None}, ctx.receiver_key
                    )

                if state is not None and receiver["state"] != state:
                    return HandlerResponse.SKIPPED

                ctx.receiver = receiver

                return await func(upd, ctx, *args, **kwargs)
            return wrapper
        return decorator



import types
from .handler import HandlerResponse
from .update import UpdateType, ReceiverType


class Context:
    """
    Use Context.create for instantiation of Context object. This object
    can store ane attributes provided by routers or other plugins.

    :ivar ~.app: Context's kutana application.
    :ivar ~.update: Update currently being processed.
    :ivar ~.backend: Backend who created update.
    :ivar ~.config: Application's configuration
    :ivar ~.default_target_id: Default receiver of responses
    """

    def __init__(self, app=None, config=None, update=None, backend=None):
        self.app = app
        self.update = update
        self.backend = backend
        self.config = config

        self.COMPLETE = HandlerResponse.COMPLETE
        self.SKIPPED = HandlerResponse.SKIPPED

        self.sender_key = None
        self.receiver_key = None
        self.default_target_id = None

    @staticmethod
    async def create(app=None, config=None, update=None, backend=None):
        """Create context with specified parameters."""

        ctx = Context(
            app=app,
            config=config,
            update=update,
            backend=backend,
        )

        if update.type == UpdateType.MSG:
            if update.receiver_type == ReceiverType.MULTI:
                ctx.default_target_id = update.receiver_id
            else:
                ctx.default_target_id = update.sender_id

            ctx.sender_key = ctx.get_key_for(sender_id=update.sender_id)
            ctx.receiver_key = ctx.get_key_for(receiver_id=update.receiver_id)
            ctx.sender_here_key = ctx.get_key_for(sender_id=update.sender_id, receiver_id=update.receiver_id)

        return ctx

    def get_key_for(self, sender_id=None, receiver_id=None):
        sender_segment = f":s{sender_id}" if sender_id else ""
        receiver_segment = f":r{receiver_id}" if receiver_id else ""
        return f"{self.backend.get_identity()}{sender_segment}{receiver_segment}"

    def get(self, name, default=None):
        return getattr(self, name, default)

    def replace_method(self, method_name, method):
        """
        Replaces this instance's method with specified name with passed method.

        Example:

        .. code-block:: python

            async def replacement(self, message, attachments=(), **kwargs):
                print(message)
            ctx = Context()
            ctx.replace_method("reply", replacement)
            await ctx.reply("hey")  # will print "hey"

        .. note::

            Many methods are coroutine, so you should replace coroutines with
            coroutines and normal functions with functions.

        :param method_name: method's name to replace
        :param method: method to replace with
        """

        if not callable(getattr(self, method_name, None)):
            raise ValueError(f"No method with name '{method_name}'")

        setattr(self, method_name, types.MethodType(method, self))

    @staticmethod
    def split_large_text(text, length=4096):
        """
        Split text into chunks with specified length.

        :param text: text for splitting
        :param length: maximum size for one chunk
        :returns: tuple of chunks
        """

        text = str(text)

        yield text[0: length]

        for i in range(length, len(text), length):
            yield text[i: i + length]

    async def request(self, method, **kwargs):
        """Perform request to backend's API and return response."""

        return await self.backend.execute_request(
            method,
            kwargs,
        )

    async def reply(self, message, attachments=(), **kwargs):
        """
        Reply with specified message and attachments. Other kwargs are
        passed to the backend. Often as additional arguments to sending the
        message.

        If message is too long - it will be splitted.
        """

        return await self.send_message(
            self.default_target_id,
            message,
            attachments,
            **kwargs
        )

    async def send_message(self, target_id, message, attachments=(), **kwargs):
        """
        Send message to specified target_id with specifiend message and
        attachments. Other kwargs are passed to the backend. Often as
        additional arguments to sending the message.

        If message is too long - it will be splitted.
        """

        responses = []

        parts = tuple(self.split_large_text(message))

        for part in parts[:-1]:
            responses.append(await self.backend.execute_send(
                target_id,
                part,
                (),
                {},
            ))

        responses.append(await self.backend.execute_send(
            target_id,
            parts[-1],
            attachments,
            kwargs,
        ))

        return responses



import os
import os.path
import yaml
from .pluralization import Pluralization


DEFAULT_LANGUAGE = "en"
TRANSLATIONS = {}


class Translation:
    __slots__ = ("msgid", "msgctx", "msgstr", "language")

    def __init__(self, msgid, msgctx, msgstr=None):
        self.msgid = msgid
        self.msgctx = msgctx
        self.msgstr = msgstr or msgid

    @staticmethod
    def make_key(msgid, msgctx):
        return (msgid, msgctx or "")

    @property
    def key(self):
        return self.make_key(self.msgid, self.msgctx)

    def get(self, num=None, language=None):
        if num is None:
            if isinstance(self.msgstr, (list, tuple)):
                raise ValueError('Translated string requires "num" argument')
            return self.msgstr
        return self.msgstr[Pluralization(language or DEFAULT_LANGUAGE).get_plural(num) % len(self.msgstr)]


def load_translations(path):
    if not path:
        return

    if os.path.isfile(path):
        load_translations_file(path)

    if not os.path.isdir(path):
        return

    for dirpath, __, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith(".yml"):
                load_translations_file(os.path.join(dirpath, filename))


def load_translations_file(path):
    if not os.path.isfile(path):
        return

    language = "".join(os.path.basename(path).rsplit(".yml", 1))

    with open(path, "r") as fh:
        strings = yaml.safe_load(fh.read())

    if language not in TRANSLATIONS:
        TRANSLATIONS[language] = {}

    for string in strings:
        translation = Translation(
            msgid=string["msgid"],
            msgctx=string.get("msgctx"),
            msgstr=string.get("msgstr"),
        )

        TRANSLATIONS[language][translation.key] = translation


def clear_translations():
    for language in TRANSLATIONS.values():
        language.clear()
    TRANSLATIONS.clear()


def set_default_language(language):
    global DEFAULT_LANGUAGE
    DEFAULT_LANGUAGE = language


def t(msgid, *args, ctx=None, num=None, lang=None, **kwargs):
    translation = Translation(msgid, ctx)
    if num is not None:
        translation.msgstr = [translation.msgstr]

    language = lang or DEFAULT_LANGUAGE
    if language in TRANSLATIONS:
        translation = TRANSLATIONS[language].get(translation.key, translation)

    return translation.get(num=num, language=language).format(*args, **kwargs)


load_translations(os.path.join(__file__, "default"))



class Pluralization:
    class NPLURALS:
        """Namespace for nplurals for languages"""

        ru = 3
        uk = ru
        en = 2

    class PLURAL:
        """Namespace for plurals for languages"""

        @staticmethod
        def ru(n):
            if n % 100 in [11, 12, 13, 14]:
                return 2
            if n % 10 == 1:
                return 0
            if n % 10 in [2, 3, 4]:
                return 1
            return 2

        uk = ru

        @staticmethod
        def en(n):
            return 0 if n == 1 else 1

    @classmethod
    def get_languages(cls):
        return [k for k in cls.NPLURALS.__dict__.keys() if k[:1] != "_"]

    def __init__(self, language):
        self.language = language

    def get_nplurals(self):
        return getattr(self.NPLURALS, self.language)

    def get_plural(self, n):
        return getattr(self.PLURAL, self.language)(n)



import json
import sqlite3
import threading
from contextlib import contextmanager
from ..storage import Storage, OptimisticLockException


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class SqliteStorage(Storage):
    """
    Storage implementation of the storage that uses sqlite3.
    """

    def __init__(self, path):
        self._path = path
        self._lock = threading.Lock()
        self.connection = None

    async def init(self):
        self.connection = sqlite3.connect(self._path, check_same_thread=False)
        self.connection.row_factory = dict_factory

        with self.cursor() as cur:
            cur.execute("""
                CREATE TABLE IF NOT EXISTS kvs (
                    key text NOT NULL,
                    val text NOT NULL,
                    ver integer NOT NULL default 0
                )
            """)
            cur.execute("CREATE UNIQUE INDEX IF NOT EXISTS kvs__key ON kvs (key)")
            cur.execute("CREATE INDEX IF NOT EXISTS kvs__key__ver ON kvs (key, ver)")

    @contextmanager
    def cursor(self):
        with self._lock:
            with self.connection:
                yield self.connection.cursor()

    async def _put(self, key, values, version=None):
        old_version = version or 0
        new_version = old_version + 1
        dumped_values = json.dumps(values, ensure_ascii=False)

        try:
            with self.cursor() as cur:
                if version:
                    cur.execute(
                        "UPDATE kvs SET val = ?, ver = ? WHERE key = ? AND ver = ?",
                        (dumped_values, new_version, key, old_version)
                    )

                if not version or cur.rowcount < 1:
                    cur.execute(
                        "INSERT INTO kvs (key, val, ver) VALUES (?, ?, ?)",
                        (key, dumped_values, new_version)
                    )
        except sqlite3.IntegrityError:
            raise OptimisticLockException(f"Failed to set values for key {key} (mismatched version)")

        return new_version

    async def _get(self, key):
        with self.cursor() as cur:
            cur.execute("SELECT * FROM kvs WHERE key = ?", (key,))
            row = cur.fetchone()
            if row:
                return {**json.loads(row["val"]), "_version": row["ver"]}
            return row

    async def _delete(self, key):
        with self.cursor() as cur:
            cur.execute("DELETE FROM kvs WHERE key = ?", (key,))



from .extensions import VkontaktePluginExtension
from .longpoll import VkontakteLongpoll
from .callback import VkontakteCallback


Vkontakte = VkontakteLongpoll  # Alias for backward compatibility


__all__ = [
    "VkontaktePluginExtension",
    "Vkontakte", "VkontakteLongpoll",
    "VkontakteCallback",
]



import pymongo
from motor.motor_asyncio import AsyncIOMotorClient
from ..storage import Storage, OptimisticLockException


class MongoDBStorage(Storage):
    """
    Storage implementation of the storage that uses running MongoDB server.
    """

    def __init__(self, mongodb_uri, database="kutana", collection="storage"):
        self._config = {"host": mongodb_uri, "database": database, "collection": collection}
        self.client = None
        self.database = None
        self.collection = None

    async def init(self):
        self.client = AsyncIOMotorClient(self._config["host"])
        self.database = self.client[self._config["database"]]
        self.collection = self.database[self._config["collection"]]
        await self.collection.create_index([("_key", pymongo.ASCENDING)], unique=True)

    async def _put(self, key, values, version=None):
        new_version = (version or 0) + 1

        try:
            await self.collection.update_one(
                {"_key": key, "_version": version or 0},
                {"$set": {
                    **values,
                    "_key": key,
                    "_version": new_version,
                }},
                upsert=True
            )
        except pymongo.errors.DuplicateKeyError:
            raise OptimisticLockException(f"Failed to set values for key {key} (mismatched version)")

        return new_version

    async def _get(self, key):
        return await self.collection.find_one({"_key": key}, projection={"_key": 0, "_id": 0})

    async def _delete(self, key):
        await self.collection.delete_one({"_key": key})



import json
import warnings
from ...helpers import uniq_by, pick
from ...handler import Handler
from ...routers import MapRouter
from ...update import UpdateType


class PayloadRouter(MapRouter):
    __slots__ = ("possible_key_sets",)

    def __init__(self, priority=7):
        """Base priority is 7"""
        super().__init__(priority=priority)
        self.possible_key_sets = []

    def _update_key_sets(self, obj):
        if not isinstance(obj, dict):
            return

        self.possible_key_sets = uniq_by([
            *self.possible_key_sets,
            list(sorted(obj.keys()))
        ], self._to_hashable)

    def _to_hashable(self, obj):
        if isinstance(obj, dict):
            return tuple(
                (k, self._to_hashable(v)) for k, v in sorted(obj.items())
            )
        if isinstance(obj, list):
            return tuple(self._to_hashable(o) for o in obj)
        return obj

    def _get_keys(self, update, ctx):
        backend_identity = ctx.backend.get_identity()

        if update.type != UpdateType.MSG or backend_identity != "vkontakte":
            return

        message = update.raw["object"]["message"]

        try:
            payload = json.loads(message.get("payload", ""))
        except json.JSONDecodeError:
            return

        if isinstance(payload, dict):
            for key_set in self.possible_key_sets:
                yield self._to_hashable(pick(payload, key_set))
        else:
            yield self._to_hashable(payload)

    def add_handler(self, handler, key):
        self._update_key_sets(key)
        return super().add_handler(handler, self._to_hashable(key))


class ActionMessageRouter(MapRouter):
    __slots__ = ()

    def __init__(self, priority=3):
        """Base priority is 3."""
        super().__init__(priority)

    def add_handler(self, handler, key):
        return super().add_handler(handler, key.lower())

    def _get_keys(self, update, ctx):
        backend_identity = ctx.backend.get_identity()

        if update.type != UpdateType.MSG or backend_identity != "vkontakte":
            return ()

        message = update.raw["object"]["message"]
        if "action" not in message:
            return ()

        action = message["action"]

        ctx.action_type = action["type"]
        ctx.action = action

        return (action["type"],)


class VkontaktePluginExtension:
    def __init__(self, plugin):
        self.plugin = plugin

    def on_payload(self, *args, **kwargs):
        warnings.warn(
            '"on_payload" is deprecated, use "on_payloads" instead',
            DeprecationWarning
        )
        return self.on_payloads(self, *args, **kwargs)

    def on_payloads(
        self,
        payloads,
        priority=0,
        router_priority=None,
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message and payload have specified
        content. Excessive fields in objects are ignored. Use
        strings and numbers for exact matching.

        Context is automatically populated with following values:

        - payload

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        def decorator(func):
            async def wrapper(update, ctx):
                ctx.payload = json.loads(
                    update.raw["object"]["message"].get("payload", "")
                )

                return await func(update, ctx)

            for payload in payloads:
                self.plugin._add_handler_for_router(
                    PayloadRouter,
                    handler=Handler(wrapper, priority),
                    handler_key=payload,
                    router_priority=router_priority,
                )

            return func

        return decorator

    def on_message_actions(
            self,
            action_types,
            priority=0,
            router_priority=None,
    ):
        """
        Decorator for registering coroutine to be called when
        incoming update is message with action (only for conversations).

        Context is automatically populated with following values:

        - action_type
        - action

        See :class:`kutana.plugin.Plugin.on_commands` for details
        about 'priority' and 'router_priority'.
        """

        def decorator(func):
            for action_type in action_types:
                self.plugin._add_handler_for_router(
                    ActionMessageRouter,
                    handler=Handler(func, priority),
                    handler_key=action_type,
                    router_priority=router_priority,
                )
            return func

        return decorator



import os
import logging
import argparse
import yaml
from kutana import Kutana, load_plugins, logger
from kutana.i18n import load_translations, set_default_language
from kutana.backends import Vkontakte, VkontakteCallback, Telegram
from kutana.storages import MemoryStorage, MongoDBStorage, SqliteStorage


parser = argparse.ArgumentParser("kutana", description="Run kutana application instance using provided config.")
parser.add_argument(
    "--config", dest="config", type=str,
    default="config.yml", help="file with config in yaml format (default: config.yml",
)
parser.add_argument(
    "--plugins", dest="plugins", type=str,
    default="plugins", help="folder with plugins to load (default: plugins)",
)
parser.add_argument(
    "--translations", dest="translations", type=str,
    default="", help="folder with translations to load (default: none)",
)
parser.add_argument(
    "--debug", dest="debug", action="store_const",
    const=True, default=False,
    help="set logging level to debug",
)


def add_backends(app, backends):
    for backend in backends or []:
        kwargs = {k: v for k, v in backend.items() if k != "kind"}
        name = kwargs.get("name")

        if name and app.get_backend(name):
            logger.logger.warning(f"Duplicated backend name: {name}")

        if backend["kind"] == "vk" and "address" in backend:
            app.add_backend(VkontakteCallback(**kwargs))

        elif backend["kind"] == "vk":
            app.add_backend(Vkontakte(**kwargs))

        elif backend["kind"] == "tg":
            app.add_backend(Telegram(**kwargs))

        else:
            logger.logger.warning(f"Unknown backend kind: {backend['kind']}")


def add_storages(app, storages):
    for storage in storages or []:
        kwargs = {k: v for k, v in storage.items() if k not in ("kind", "name")}
        name = storage.get("name", "default")

        if name != "default" and app.get_storage(name):
            logger.logger.warning(f"Duplicated storage name: {name}")

        if storage["kind"] == "memory":
            app.set_storage(name, MemoryStorage(**kwargs))

        elif storage["kind"] == "mongodb":
            app.set_storage(name, MongoDBStorage(**kwargs))

        elif storage["kind"] == "sqlite":
            app.set_storage(name, SqliteStorage(**kwargs))

        else:
            logger.logger.warning(f"Unknown storage kind: {storage['kind']}")


def run():
    """
    This function runs kutana application using provided
    configuration and CLI options.

    Refer to its source to create more specific starter for
    your application.
    """

    # Parse provided arguments
    args = parser.parse_args()

    # Setup logger
    if args.debug:
        logger.set_logger_level(logging.DEBUG)

    # Import configuration
    if not os.path.isfile(args.config):
        logger.logger.error(f"Couldn't open confiuration file: {args.config}")
        exit(1)

    with open(args.config) as fh:
        config = yaml.safe_load(fh)

    # Create application
    app = Kutana()

    # Update configuration
    app.config.update(config)

    # Setup i18n
    set_default_language(config.get("language", "en"))
    load_translations(args.translations)

    # Add each backend from config
    add_backends(app, config.get("backends"))

    # Add each storage from config
    add_storages(app, config.get("storages"))

    # Load and register plugins from provided path
    app.add_plugins(load_plugins(args.plugins))

    # Run application
    app.run()


if __name__ == "__main__":
    run()



import asyncio
from sortedcontainers import SortedList
from .handler import HandlerResponse as hr
from .storages import MemoryStorage
from .storage import OptimisticLockException, Storage
from .backend import Backend
from .context import Context
from .plugin import Plugin
from .logger import logger


class Kutana:
    """
    Main class for kutana application

    Configuration:

    - '.prefixes' - prefixes for commands (default is [".", "/"])
    - '.ignore_initial_spaces' - ignore spaces after prefix (default is True)

    :ivar ~.config: Application's configuration
    """

    def __init__(
        self,
        concurrent_handlers_count=512,
        default_storage=None,
        loop=None,
    ):
        self._plugins = []
        self._backends = []
        self._storages = {"default": default_storage or MemoryStorage()}

        self._loop = loop or asyncio.new_event_loop()

        self._concurrent_handlers_count = concurrent_handlers_count
        self._sem = asyncio.Semaphore(
            value=concurrent_handlers_count, loop=self._loop
        )

        self._routers = None
        self._handlers = None

        self.config = {
            "prefixes": (".", "/"),
            "ignore_initial_spaces": True
        }

    def get_loop(self):
        """Return application's asyncio loop."""
        # TODO: Replace with property
        return self._loop

    def set_storage(self, name, storage):
        if not isinstance(storage, Storage):
            raise ValueError(f"Provided value is not a storage: {storage}")
        self._storages[name] = storage

    def get_storage(self, name="default"):
        return self._storages.get(name)

    def add_plugin(self, plugin):
        """Add plugin to the application."""
        if not isinstance(plugin, Plugin):
            raise ValueError(f"Provided value is not a plugin: {plugin}")
        if plugin in self._plugins:
            raise RuntimeError("Plugin already added")
        self._plugins.append(plugin)

    def add_plugins(self, plugins):
        """Add every plugin in passed iterable to the application."""
        for plugin in plugins:
            self.add_plugin(plugin)

    def get_plugins(self):
        """Return list of added plugins."""
        return self._plugins

    def add_backend(self, backend):
        """Add backend to the application."""
        if not isinstance(backend, Backend):
            raise ValueError(f"Provided value is not a backend: {backend}")
        if backend in self._backends:
            raise RuntimeError("Backend already added")
        self._backends.append(backend)

    def get_backend(self, name):
        """Return first backend with specified name or None."""
        for backend in self._backends:
            if backend.name == name:
                return backend
        return None

    def get_backends(self):
        return self._backends

    async def _on_start(self, queue):
        for storage in self._storages.values():
            await storage.init()

        # Prepare backends and run background update acquiring
        for backend in self._backends:
            await backend.on_start(self)

            if not backend.active:
                continue

            async def acquire_updates(backend):  # don't forget to capture backend
                async def submit_update(update):
                    return await queue.put((update, backend))

                while True:
                    if queue.qsize() < queue.maxsize:
                        await backend.acquire_updates(submit_update)
                    await asyncio.sleep(0)

            asyncio.ensure_future(acquire_updates(backend), loop=self._loop)

        # Prepare plugins
        for plugin in self._plugins:
            plugin.app = self

        # Run event listeners
        await self._handle_event("start")

    async def _main_loop_wrapper(self):
        try:
            await self._main_loop()
        except asyncio.CancelledError:
            pass
        except Exception:
            self.stop()
            raise

    async def _main_loop(self):
        queue = asyncio.Queue(
            maxsize=self._concurrent_handlers_count,
            loop=self._loop
        )

        await self._on_start(queue)

        while True:
            await self._sem.acquire()

            update, backend = await queue.get()

            ctx = await Context.create(
                app=self,
                config=self.config,
                update=update,
                backend=backend,
            )

            backend.prepare_context(ctx)

            task = asyncio.ensure_future(
                self._handle_update_with_logger(update, ctx),
                loop=self._loop
            )

            task.add_done_callback(lambda t: self._sem.release())

    def _init_handlers(self):
        self._handlers = {}

        for event in ["start", "before", "after", "exception", "shutdown"]:
            self._handlers[event] = SortedList([], key=lambda r: -r.priority)

        for plugin in self._plugins:
            for event, handlers in plugin._handlers.items():
                self._handlers[event].update(handlers)

    async def _handle_event(self, name, *args, **kwargs):
        if self._handlers is None:
            self._init_handlers()

        for handler in self._handlers.get(name):
            await handler.handle(*args, **kwargs)

    def _init_routers(self):
        self._routers = SortedList([], key=lambda r: -r.priority)

        def _add_router(new_router):
            for router in self._routers:
                if router.alike(new_router):
                    router.merge(new_router)
                    return

            self._routers.add(new_router)

        for plugin in self._plugins:
            for router in plugin._routers:
                _add_router(router)

    async def _handle_update_with_logger(self, update, ctx):
        logger.debug("Processing update %s", update)

        try:
            return await self._handle_update(update, ctx)
        except asyncio.CancelledError:
            pass
        except OptimisticLockException as exc:
            logger.debug("Optimistic lock exception: %s", exc)
        except Exception as exc:
            logger.exception("Exception while handling the update")
            await self._handle_event("exception", update, ctx, exc)

    async def _handle_update(self, update, ctx):
        if self._routers is None:
            self._init_routers()

        await self._handle_event("before", update, ctx)

        for router in self._routers:
            if await router.handle(update, ctx) != hr.SKIPPED:
                ctx._result = hr.COMPLETE
                break
        else:
            ctx._result = hr.SKIPPED

        await self._handle_event("after", update, ctx, ctx._result)

        return ctx._result

    async def _shutdown(self):
        logger.info("Gracecfully shutting application down...")

        # Clean up
        tasks = []

        for backend in self._backends:
            tasks.append(backend.on_shutdown(self))

        tasks.append(self._handle_event("shutdown"))

        await asyncio.gather(*tasks, loop=self._loop, return_exceptions=True)

        # Cancel everything else
        tasks = []

        for task in asyncio.Task.all_tasks(loop=self._loop):
            if task is not asyncio.Task.current_task():
                task.cancel()
                tasks.append(task)

        await asyncio.gather(*tasks, loop=self._loop, return_exceptions=True)

        self._loop.stop()

    def run(self):
        """Run the application."""
        logger.info("Starting application...")

        try:
            asyncio.ensure_future(self._main_loop_wrapper(), loop=self._loop)
            self._loop.run_forever()
        except KeyboardInterrupt:
            asyncio.ensure_future(self._shutdown(), loop=self._loop)
            self._loop.run_forever()
        finally:
            self._loop.close()
            logger.info("Stopped application")

    def stop(self):
        """Stop the application."""
        return asyncio.ensure_future(self._shutdown(), loop=self._loop)



class Backend:
    def __init__(self, name=None, active=True):
        self.name = name
        self._active = active

    @property
    def active(self):
        return self._active

    async def on_start(self, app):
        pass

    async def on_shutdown(self, app):
        pass

    def prepare_context(self, ctx):
        pass

    async def acquire_updates(self, submit_update):
        raise NotImplementedError

    async def execute_send(self, target_id, message, attachments, kwargs):
        raise NotImplementedError

    async def execute_request(self, method, kwargs):
        raise NotImplementedError

    @classmethod
    def get_identity(cls):
        return cls.__name__.lower()
